(require 'eieio)

;; class for single phrase
(defclass vocab-entry ()
  ((native :initarg :native :initform "" :type string
	   :documentation "phrase in native lang")
   (foreign :initarg :foreign :initform "" :type string
	    :documentation "phrase in foreign lang")
   (score :initarg :score :initform 0 :type integer
	  :documentation "score")))

;; class for trainer
(defclass vocab-trainer ()
  ((entries :initarg :entries :initform nil :type list
	    :documentation "list of phrases")))

(defmethod load-vocab ((trainer vocab-trainer))
  "Reading phrases from current buffer and saving as objecst."
  (with-current-buffer (current-buffer)
    (oset trainer entries nil)
    (goto-char (point-min))
    (while (not (eobp))
      (let* ((line (thing-at-point 'line t))
             (parts (split-string line " | " t))
             (native (nth 0 parts))
             (foreign (nth 1 parts))
             (score (string-to-number (nth 2 parts))))
        (push (vocab-entry :native native :foreign foreign :score score)
	      (oref trainer entries)))
      (forward-line 1))))

(defmethod save-vocab ((trainer vocab-trainer))
  "Saving phrases with actual scores back to current buffer."
  (with-current-buffer (current-buffer)
    (erase-buffer)
    (let ((sorted_entries (cl-sort (copy-sequence (oref trainer entries)) '<
		      :key (lambda (e) (oref e score)))))
      (dolist (entry sorted_entries)
        (insert (format "%s | %s | %d\n"
                        (oref entry native)
                        (oref entry foreign)
                        (oref entry score)))))
    (save-buffer)))

(defmethod select-weakest ((trainer vocab-trainer))
  "Selecting phrases with lowest score."
  (cl-subseq (cl-sort (copy-sequence (oref trainer entries)) '<
		      :key (lambda (e) (oref e score))) 
             0 12))

(defmethod train ((trainer vocab-trainer))
  "Start training."
  (let ((words (select-weakest trainer)))
    (while words
      (let* ((entry (car words))
             (user-input (read-string (format "Translate: %s → " (oref entry native)))))
        (if (string= user-input (oref entry foreign))
            (progn
              (message "Correct!")
	      (sit-for 1)
              (oset entry score (1+ (oref entry score)))
              (setq words (cdr words)))
          (progn
	    (message "Wrong! Correct answer: %s (press key...)"
		     (oref entry foreign))
	    (read-char)
            (oset entry score (1- (oref entry score))))))))
  (save-vocab trainer))

;; Launch trainer.
(defun start-vocab-trainer ()
  (interactive)
  (let ((trainer (vocab-trainer)))
    (load-vocab trainer)
    (train trainer)))
